"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("fs");
const http = require("http");
const opentracing_1 = require("opentracing");
const os = require("os");
require("reflect-metadata");
const util_1 = require("util");
const brokers_1 = require("./brokers");
const Error_1 = require("./Error");
const Fork_1 = require("./Fork");
const ForkHttp_1 = require("./ForkHttp");
const types_1 = require("./types");
const Worker_1 = require("./Worker");
const stringy_1 = require("stringy");
const UdpClient_1 = require("./UdpClient");
class Microservice extends brokers_1.default {
    /**
     * Contructor de la clase
     * @param conf Configuracion del servicio
     */
    constructor(conf) {
        // Llamar el constructor del padre
        super(conf);
        /**
         * Coleccion de errores
         * @private
         * @type Map
         */
        this.ErrorColl = new Map();
        /**
         * Coleccion de errores custom
         * @private
         * @type Map
         */
        this.ErrorCustomColl = new Map();
        /**
         * Coleccion de Forks (bifurcaciones)
         * @private
         * @type Map
         */
        this.ForksColl = new Map();
        /**
         * Hostname del nodo
         * @private
         */
        this.hostname = os.hostname();
        // Identificar si es NATS
        if (process.env.NATS) {
            conf.IsNats = true;
        }
        else {
            conf.IsNats = false;
        }
        // Errores por default
        const errDefault = {
            '-96': 'No se indico la direccion del endpoint para el consumo de la bifurcacion',
            '-97': 'Error en revision de duplicados',
            '-98': 'Ocurrio una error, favor de intentar de nuevo.',
            '-99': 'Ocurrio una excepcion, favor de intentar de nuevo.',
            '0': 'Smoke Test',
            '1': 'Este servicio no puede ser consumido de forma externa',
        };
        // Errores
        let err;
        // Validar si se definieron Forks para indicar que es bifurcacion
        if (Reflect.hasMetadata('definition:fork', this)) {
            // Indicar que es un Fork
            this.conf.Fork = true;
            // Indicar que esta en modo bifurcacion
            if (this.conf.Debug) {
                this.logs.info('Se inicia el worker con soporte para Bifurcaciones.');
            }
            // Validar que se pase la variable de URL_ENDPOINT_SERVICES
            if (process.env.REST_ENABLED) {
                if (!process.env.URL_ENDPOINT_SERVICES) {
                    this.logs.error('No se proporciono la variable URL_ENDPOINT_SERVICES');
                    process.exit(-1);
                }
            }
        }
        this.frameworkVersion = `ts_${JSON.parse(fs.readFileSync(`${__dirname}/../package.json`).toString()).version}`;
        // Agregar los Forks
        if (this.conf.Fork) {
            // Recuperar la funcion
            const fnc = this.get(Reflect.getMetadata('definition:fork', this).key);
            // Registrar
            const forks = fnc();
            // Recorrer para registrar
            for (const key of Object.keys(forks)) {
                // Almacenar
                this.ForksColl.set(key, forks[key]);
            }
        }
        // Revisar si esta definido los errores.
        if (Reflect.hasMetadata('definition:errors', this)) {
            // Recuperar la funcion
            const fnc = this.get(Reflect.getMetadata('definition:errors', this).key);
            // Asignar
            err = Object.assign(fnc(), errDefault);
        }
        else {
            // Asignar
            err = errDefault;
        }
        // Almacenar
        for (const key of Object.keys(err)) {
            // Almacenar el Error
            this.ErrorColl.set(parseInt(key, 0), err[key]);
        }
        // Iniciar Healthcheck
        this.startHttpHealthCheck();
    }
    /**
     * Metodo para el procesamiento de mensajes
     * @param data mensaje
     */
    message(data, callback) {
        return __awaiter(this, void 0, void 0, function* () {
            let isSmoketest = false;
            // FIX: crear el objeto HTTP en caso de que no exista
            if (!data.metadata.hasOwnProperty('http')) {
                data.metadata = Object.assign(data.metadata, {
                    http: {
                        code: 200,
                        headers: null,
                        message: '',
                    },
                });
            }
            // Metodo que se ejecutara
            let methondExecute = null;
            // Accion que se esta ejecutando
            let actionName = null;
            // Validar smoketest
            try {
                if (data.hasOwnProperty('data')) {
                    if (data.data.hasOwnProperty('smoketest')) {
                        isSmoketest = true;
                    }
                    else if (data.hasOwnProperty('metadata')) {
                        if (data.metadata.hasOwnProperty('filters')) {
                            if ('smoketest' in data.metadata.filters) {
                                isSmoketest = true;
                            }
                        }
                    }
                }
            }
            catch (error) {
                this.logs.error(error);
            }
            // Validar si no es una peticion de SmokeTest
            if (isSmoketest) {
                this.logs.info('Ejecutando las pruebas de smoketest');
                // Ejecutar la prueba
                try {
                    if (yield this.smoketest()) {
                        // Retornar el valor originar
                        this.response(data, 0, null, true, callback);
                    }
                    else {
                        // Pasar a enviar el error
                        this.response(data, -1, null, true, callback);
                    }
                }
                catch (error) {
                    // Generar el transactionId
                    const transactionId = new Date().getTime();
                    // Indicar el error
                    this.sendError(this.Topic, transactionId, error);
                    // Pasar a enviar el error
                    this.response(data, -1, null, true, callback);
                }
            }
            else {
                // Ver si esta activo el modo Debug
                if (this.conf.Debug) {
                    // Imprimir el contenido
                    this.logs.info(`
                    ENTRADA [${data.metadata.owner}]:
                    HEADERS: [${JSON.stringify(data.headers)}]
                    DATA: [${JSON.stringify(data.data)}]
                `);
                }
                // Validar si tiene asignado la metadata original
                if (!('metadata_inicial' in data.metadata)) {
                    data.metadata.metadata_inicial = Object.assign({}, data.metadata);
                }
                // Validar si tiene almacenada los datos originales
                if (!('data_inicial' in data.metadata)) {
                    data.metadata.data_inicial = Object.assign({}, data.data);
                }
                if (data.metadata.time.indexOf('Z') > -1 && this.conf.ClearQueque) {
                    // Cast de la fecha
                    const date = new Date(data.metadata.time);
                    // Fecha actual
                    const currDate = new Date();
                    // Sacar la diferencia de fechas
                    const diff = Math.round((currDate.getTime() - date.getTime()) / 60000);
                    // Si pasaron 5 minutos ignorar el mensaje
                    if (diff >= 5) {
                        return;
                    }
                }
                // Ver sis es una bifurcacion para cambiar el metodo.
                if (data.metadata.bifurcacion &&
                    this.Topic === data.metadata.metadata_inicial.owner) {
                    // Indicar el metodo heredado
                    data.metadata.method = data.metadata.metadata_inicial.method;
                    // Indicar el uuid original.
                    data.metadata.uuid = data.metadata.metadata_inicial.uuid;
                }
                // Validar que cuente con el uuid y method
                if (data.metadata.method &&
                    data.metadata.uuid &&
                    data.metadata.uuid !== '') {
                    if (data.metadata.method === types_1.HttpMethod.PUT) {
                        actionName = 'definition:update';
                        // Buscar si cuenta con una definicion de actualizacion
                        methondExecute = this.getConfHandler(actionName);
                    }
                    else if (data.metadata.method === types_1.HttpMethod.DELETE) {
                        actionName = 'definition:delete';
                        // Buscar el metodo de eliminacion
                        methondExecute = this.getConfHandler(actionName);
                    }
                    else if (data.metadata.method === types_1.HttpMethod.GET) {
                        actionName = 'definition:get';
                        // Buscar el metodo de consulta
                        methondExecute = this.getConfHandler(actionName);
                    }
                }
                else if (data.metadata.method &&
                    data.metadata.method === types_1.HttpMethod.GET) {
                    actionName = 'definition:list';
                    // Buscar el metodo de List
                    methondExecute = this.getConfHandler(actionName);
                }
                else if (data.metadata.method &&
                    data.metadata.method === types_1.HttpMethod.POST) {
                    actionName = 'definition:create';
                    // Buscar si cuenta con una definicion de registro
                    methondExecute = this.getConfHandler(actionName);
                }
                else {
                    actionName = 'definition:listener';
                    // Buscar le metodo de listener
                    methondExecute = this.getConfHandler('');
                }
                // Verficar si no ocurrio un error al consultar la funcion
                if (util_1.isNull(methondExecute)) {
                    return 0;
                }
                // Resolver el nombre de la defincion del metodo
                actionName = this.resolveDefinition(actionName);
                // Se revisa el span del padre y si lo tiene se asigna en la configuracion
                const rootSpan = data.metadata.span;
                let span;
                if (rootSpan) {
                    let spanIdOrObject = null;
                    // Validar si es un string o un objecto
                    /**
                     * ==========================================
                     * |    REVISAR POR QUE DEJO DE FUNCIONAR
                     * ==========================================
                     */
                    if (rootSpan instanceof Object) {
                        spanIdOrObject = rootSpan;
                    }
                    else {
                        spanIdOrObject = this.tracer.extract(opentracing_1.FORMAT_TEXT_MAP, rootSpan);
                    }
                    const configSpan = {
                        childOf: spanIdOrObject,
                    };
                    span = this.tracer.startSpan(`${methondExecute.key}`, configSpan);
                }
                else {
                    span = this.tracer.startSpan(`${methondExecute.key}`);
                }
                span.setTag(opentracing_1.Tags.HTTP_METHOD, data.metadata.method);
                span.setTag('transactionId', data.metadata.id_transaction);
                // Respuesta de ejecutar la funcion
                let resp;
                // Pasar el mensaje para ser procesado
                try {
                    if (this.conf.Fork) {
                        // Es un Fork/Bifurcacion
                        let conf = null;
                        try {
                            conf = this.getConfForks(actionName);
                        }
                        catch (error) {
                            this.logs.error(error);
                        }
                        // Configurar
                        if (conf) {
                            if (process.env.REST_ENABLED) {
                                // ForkHttp
                                this.app = new ForkHttp_1.default(this.conf, this.Topic, span, this.tracer);
                            }
                            else {
                                // Si tiene configuracion este metodo se cambia a Fork
                                this.app = new Fork_1.default(this.conf, this.Topic, span);
                            }
                            this.app.forksConfig(conf);
                        }
                        else {
                            // Si no tiene configuracion este metodo se cambia a Worker
                            this.app = new Worker_1.default(this.conf, null, span);
                        }
                    }
                    else {
                        this.app = new Worker_1.default(this.conf, null, span);
                    }
                    resp = yield this.app.message(data, this.get(methondExecute.key).bind(this), this.getParams(methondExecute.key));
                }
                catch (error) {
                    this.logs.error(JSON.stringify(error.error));
                    let cache = [];
                    // Enviar a errores criticos
                    this.sendError(this.Topic, data.metadata.id_transaction, error);
                    // Enviar indicando que ocurrio un error.
                    this.response(Object.assign(JSON.parse(JSON.stringify(error.response, (key, value) => {
                        if (typeof value === 'object' &&
                            value !== null) {
                            // Duplicate reference found, discard key
                            if (cache.includes(value))
                                return;
                            // Store value in our collection
                            cache.push(value);
                        }
                        return value;
                    })), {}), Object.assign(JSON.parse(JSON.stringify(error, (key, value) => {
                        if (typeof value === 'object' &&
                            value !== null) {
                            // Duplicate reference found, discard key
                            if (cache.includes(value))
                                return;
                            // Store value in our collection
                            cache.push(value);
                        }
                        return value;
                    })), {}) || -99, span, false, callback);
                    cache = null;
                    return;
                }
                // Enviar la respuesta
                try {
                    if (util_1.isUndefined(resp)) {
                        //
                    }
                    else {
                        // Enviar la respuesta del error
                        this.response(resp.response, resp, span, false, callback);
                    }
                }
                catch (error) {
                    this.logs.error(error);
                    // Enviar a errores criticos
                    // this.sendError(this.Topic, data.metadata.id_transaction, error);
                    // Enviar indicando que ocurrio un error.
                    this.response(Object.assign(JSON.parse(JSON.stringify(error.response)), {}), Object.assign(JSON.parse(JSON.stringify(error)), {}) || -99, span, false, callback);
                }
                span.finish();
            }
        });
    }
    /**
     * Metodo para publicar un error en la cola de mensajes
     * @param service Nombre del servicio
     * @param transactionId Id de la transaccion
     * @param error Error
     */
    sendError(service, transactionId, error) {
        const dt = {
            containerID: this.hostname,
            error,
            framework: this.frameworkVersion,
            id: transactionId,
            servicio: service,
        };
        let message = null;
        // Validar si se envia por UDP o por la cola de mensajes
        if (process.env.REST_ENABLED) {
            message = Buffer.from(stringy_1.stringify(dt)).toString('base64');
            // Enviar por udp
            UdpClient_1.default.Send(Buffer.from(message), // Pasar al buffer para pasarlo por udp
            process.env.UDP_PORT
                ? process.env.UDP_PORT // Realizar la conversion
                : 10000, process.env.UDP_HOST || 'localhost');
        }
        else {
            // Asignar el objeto de respuesta
            message = Buffer.from(JSON.stringify(dt)).toString('base64');
            // Enviar el error
            this.send(message, 'Errores_criticos', transactionId);
        }
    }
    /**
     * Metodo para publicar la respuesta
     *
     * @param {DataRequest} data
     * @param {(number | ResponseMessage)} error
     * @param {Span} [span]
     * @param {boolean} [smoketest=false]
     * @memberof Microservice
     */
    response(data, error, span, smoketest = false, callback) {
        // Codigo del error.
        let errorCode = 0;
        // Bandera para conocer se proporciona error metadata
        let isForkError = false;
        // Comprovar si se proporciono un codigo de error o una metadata del error
        if (typeof error === 'number') {
            errorCode = error;
        }
        else {
            errorCode = error.error.errorCode;
            isForkError = error.fork; // indicar que es un error desde una bifurcacion.
        }
        // Recuperar los datos del error
        const instanceError = this.getError(errorCode);
        // Validar que tipo es
        const isHttpError = instanceError.hasOwnProperty('httpCode') &&
            instanceError.hasOwnProperty('message');
        // Objecto de salida
        let Resp = {};
        // Cache
        let cache = [];
        // Datos de respuesta
        const dataResponse = Object.assign({}, data);
        cache = null;
        // Error generico
        const errorGen = {
            errorCode,
            framework: this.frameworkVersion,
            userMessage: isHttpError ? instanceError.message : instanceError,
        };
        // Ver si se tiene asignado un mensaje de error perzonalizado
        if (data.metadata.http.message && data.metadata.http.message !== '') {
            errorGen.userMessage = data.metadata.http.message;
        }
        // Objecto de error
        const errorResponse = {
            data: {
                response: isForkError
                    ? Object.assign(error.error, {
                        framework: this.frameworkVersion,
                    })
                    : errorGen,
            },
            meta: {
                id_transaction: dataResponse.metadata.id_transaction,
                status: errorCode !== 0 ? 'ERROR' : 'SUCCESS',
            },
        };
        // Validar si ocurrio un error
        if (errorCode !== 0) {
            // Asignar el hostname
            errorResponse.data.response.hostname = this.hostname;
            // Codigo de error
            errorResponse.data.response.code = errorCode;
            // Eliminar errorcode
            // delete errorResponse.data.response.errorCode;
            // Asignar el error
            dataResponse.response = errorResponse;
        }
        // Asignar el tiempo
        dataResponse.metadata.time = new Date().toISOString();
        // Topico
        dataResponse.metadata.owner = this.Topic;
        // Tipo de salida
        dataResponse.metadata.mtype = 'output';
        // Hostname
        dataResponse.metadata.hostname = this.hostname;
        // Framework
        dataResponse.metadata.framework = this.frameworkVersion;
        // Validar que tenga un uowner
        if ('uowner' in data.metadata) {
            // Asignar
            dataResponse.metadata.uworker = data.metadata.uowner;
        }
        // Validar si tiene asignado un uworker
        if ('uworker' in data.metadata) {
            // Asignar
            dataResponse.metadata.uowner = data.metadata.uworker;
        }
        // Topico de respuesta
        let topic = null;
        if (span) {
            const carrier = {};
            this.tracer.inject(span, opentracing_1.FORMAT_TEXT_MAP, carrier);
            dataResponse.metadata.span = carrier;
        }
        // Si app es una instancia de Worker cambiar bifurcacion a false
        if (this.app instanceof Worker_1.default) {
            dataResponse.metadata.bifurcacion = false;
        }
        // Revisamos si es bifurcacion
        if (dataResponse.metadata.bifurcacion) {
            const regCallback = {
                method: dataResponse.metadata.original_method,
                name: this.Topic,
            };
            const transaction = dataResponse.metadata.id_transaction;
            const method = dataResponse.metadata.original_method;
            if (!dataResponse.metadata.callbacks) {
                dataResponse.metadata.callbacks = [];
            }
            // Validar que no existe previamente.
            if (this.existCallback(dataResponse.metadata.callbacks, regCallback)) {
                this.logs.error(`Loop detection:  Transaction: ${transaction} - method: ${method} - name: ${topic}`);
                // Enviar el Error a kafka
                this.sendError(topic, dataResponse.metadata.id_transaction, new Error('loop detection'));
            }
            else {
                // Se agrega asi misma a la pila
                dataResponse.metadata.callbacks.push(regCallback);
            }
        }
        else {
            // Se obtiene el ultimo elemento agregado para responder a el
            if (dataResponse.metadata.callbacks &&
                dataResponse.metadata.callbacks.length > 0) {
                let cllbs = dataResponse.metadata.callbacks.pop();
                const sizeCllbs = dataResponse.metadata.callbacks.length;
                const meta = dataResponse.metadata;
                // Se valida si es de plantilla vieja
                const originalMethod = meta.original_method
                    ? meta.original_method
                    : meta.metadata_inicial
                        ? meta.metadata_inicial.method
                        : meta.method
                            ? meta.method
                            : 'GET';
                if (this.Topic === cllbs.name &&
                    originalMethod === cllbs.method) {
                    for (let x = 0; x < sizeCllbs && !topic; x++) {
                        cllbs = dataResponse.metadata.callbacks.pop();
                        if (this.Topic !== cllbs.name ||
                            originalMethod !== cllbs.method) {
                            topic = cllbs.name;
                            dataResponse.metadata.method = cllbs.method;
                        }
                    }
                }
                else {
                    topic = cllbs.name;
                    dataResponse.metadata.method = cllbs.method;
                }
            }
        }
        // Eliminamos elemento (Este es agregado cuando es una bifurcacion)
        delete dataResponse.metadata.original_method;
        // Validar sobre cual elemento publica
        if (util_1.isUndefined(dataResponse.uuid) || util_1.isNull(dataResponse.uuid)) {
            topic = topic ? topic : `respuesta_${data.metadata.owner}`; // KAFKA
        }
        else {
            const reg = new RegExp(/[a-zA-Z0-9]{1,}[_.]{1}[a-zA-Z0-9]{1,}[_.]{1}.*/);
            topic = reg.test(topic) ? topic : dataResponse.uuid; // Id del reply NATS
        }
        // Consultar quien es el callback
        const cllb = dataResponse.metadata.callback;
        if (dataResponse.metadata.worker !== dataResponse.metadata.owner &&
            dataResponse.metadata.bifurcacion) {
            // Asignar el topico que es el Worker
            topic = dataResponse.metadata.worker;
            // Indicar el worker
            dataResponse.metadata.worker = data.metadata.owner;
            // Data original
            Resp = Object.assign(dataResponse, {});
        }
        else if (this.Topic !== cllb && dataResponse.metadata.bifurcacion) {
            // Asignar el topico
            topic = topic ? topic : cllb;
            // Asignar la data
            dataResponse.data = dataResponse.response.data;
            // Indicar los datos
            Resp = Object.assign(dataResponse, {});
        }
        else {
            // Asignar la respuesta
            Resp = {
                _id: dataResponse.metadata.id_transaction,
                data: dataResponse.metadata.data_inicial,
                headers: dataResponse.headers,
                metadata: dataResponse.metadata,
                response: Object.assign(dataResponse.response || {}, {}),
                uuid: dataResponse.uuid,
            };
        }
        // Si es el ultimo callback eliminamos la metadata y data inicial
        if (Resp.metadata.callbacks && Resp.metadata.callbacks.length === 0) {
            delete Resp.metadata.metadata_inicial;
            delete Resp.metadata.data_inicial;
        }
        // Revisar si es un smoketest
        if (smoketest) {
            // Objeto de respuesta generica
            Resp.response = {
                data: {
                    response: {
                        code: errorCode,
                        framework: this.frameworkVersion,
                        hostname: this.hostname,
                        userMessage: 'Smoke test',
                    },
                },
                meta: {
                    id_transaction: dataResponse.metadata.id_transaction,
                    status: errorCode < 0 ? 'ERROR' : 'SUCCESS',
                },
            };
        }
        else {
            span.log({ event: 'output', response: Resp.response });
        }
        // Validar si es modeo DEBUG
        if (this.conf.Debug) {
            // Imprimir por consola lo que se envia
            this.logs.info(`SALIDA [${topic}]: ${stringy_1.stringify(Resp)}`);
        }
        // Pasar a Base64 la respuesta
        const dataBuffer = Buffer.from(stringy_1.stringify(Resp)).toString('base64');
        // Validar si es callback o cola
        if (process.env.REST_ENABLED) {
            if (!util_1.isUndefined(callback)) {
                callback(Resp);
            }
        }
        else {
            // Enviar la respuesta
            this.send(dataBuffer, topic);
        }
    }
    /**
     * Metodo que se ejecuta al ocurrir una excepcion al procesar mensajes.
     * @param topic Topico del Error
     * @param data Mensaje de la peticion
     * @param error Error que se ejecuta
     */
    exception(topic, data, error) {
        // Generar el transactionId
        const transactionId = new Date().getTime();
        // Enviar el Error a kafka
        this.sendError(topic, transactionId, error);
        // Retornar el error al consumidor
        this.response({
            data: null,
            metadata: {
                id_transaction: `error_${transactionId}`,
                owner: this.Topic,
            },
        }, -99);
    }
    /**
     * Metodo para validar que la seguridad necesaria
     * sea valida.
     * @param token Token de seguridad
     */
    auth(token) {
        return new Promise((rr) => {
            rr(true);
        });
    }
    /**
     * Metodo que se ejecuta en el healthcheck
     * @publick
     */
    HealthCheck() {
        return this.validateConnection();
    }
    /**
     * Metodo para verificar si existe un codigo de error
     * ya registrado.
     * @param code Codigo del error
     */
    hasError(code) {
        return this.ErrorColl.has(code);
    }
    /**
     * Metodo para agregar nuevos errores a la coleccion.
     * @param text Texto del error
     * @param code Codigo del error
     */
    addError(text, code) {
        // Verificar que no exista
        if (!this.ErrorColl.has(code)) {
            // No existe agregarlo
            this.ErrorColl.set(code, text);
        }
    }
    /**
     * Metodo para generar una excepcion en el codigo
     * @param code Codigo de error
     */
    emitError(code, message = null, status = null, headers = {}) {
        // Recuperar los datos del error definido.
        const error = this.getError(code);
        // Mensaje que se enviara
        let msg = '';
        // Validar si un objeto
        const isHttpError = error.hasOwnProperty('httpCode') && error.hasOwnProperty('message');
        // Se le da prioridad al mensaje pasado por parametro
        if (message) {
            msg = message;
        }
        else if (isHttpError) {
            const d = error;
            const colr = d;
            msg = colr.message;
        }
        else {
            msg = error;
        }
        // Verificar si un objeto de respuesta http
        if (isHttpError) {
            const errorC = error;
            // Generar la instancia del error
            return new Error_1.MsError(msg, code, status ? status : errorC.httpCode, headers);
        }
        else {
            // Generar la instancia del error - generico
            return new Error_1.MsError(msg, code, status ? status : 500, headers);
        }
    }
    /**
     * Metodo para iniciar el HealthCheck HTTP
     * @private
     */
    startHttpHealthCheck() {
        // Crear la instancia del servidor
        const server = http.createServer((request, response) => __awaiter(this, void 0, void 0, function* () {
            // Validar
            try {
                if ((yield this.HealthCheck()) &&
                    (yield this.smoketest())) {
                    response.end('ok');
                }
                else {
                    // Cambiar la cabecera
                    response.statusCode = 500;
                    response.end('ERROR');
                }
            }
            catch (error) {
                // Cambiar la cabecera
                response.statusCode = 500;
                response.end('ERROR');
            }
        }));
        if (process.env.NODE_ENV === 'production' || process.env.PRODUCTION) {
            // Iniciar el server
            server.listen(80, () => {
                // Inidicar que todo esta bien
                this.logs.info('[HealthCheck] listening on 0.0.0.0:80');
            });
        }
    }
    /**
     * Metodo para recuperar un Error.
     * @param code Codigo de Error
     */
    getError(code) {
        if (this.ErrorColl.has(code)) {
            return this.ErrorColl.get(code);
        }
        else if (this.ErrorCustomColl.has(code)) {
            return this.ErrorCustomColl.get(code);
        }
        return this.ErrorColl.get(-99);
    }
    /**
     * Metodo para recuperar la configuracion de un elemento
     * registrado con los decoradores.
     * @param Name Nombre de la configuracion
     */
    getConfHandler(Name) {
        // Validar si existe la funcion que se solicita
        if (Reflect.hasMetadata(Name, this)) {
            // Regresar la configuracion
            return Reflect.getMetadata(Name, this);
        }
        else if (Reflect.hasMetadata('definition:listener', this)) {
            // Validar si existe listener
            // Regresar el generico
            return Reflect.getMetadata('definition:listener', this);
        }
        else {
            // No existe ningun elemento que responda a la peticion.
            return null;
        }
    }
    /**
     * Metodo para resolver la definicion que se ejecutara.
     * @param name Nombre de la definicion que se busca
     */
    resolveDefinition(name) {
        if (Reflect.hasMetadata(name, this)) {
            return name;
        }
        else {
            return 'definition:listener';
        }
    }
    /**
     * Metodo para recuperar los datos de configuracion para una
     * accion.
     * @param NameAction Nombre de la accion
     */
    getConfForks(NameAction) {
        // Configuracion de la bifurcacion.
        let configFork;
        // Validar que exista el metodo.
        if (Reflect.hasMetadata(`${NameAction}:conf`, this)) {
            configFork = JSON.parse(JSON.stringify(Reflect.getMetadata(`${NameAction}:conf`, this)));
        }
        else {
            // Elemento por default
            return null;
        }
        // Recuperar los datos de configuracion
        if (configFork) {
            // Recuperar los datos de configuracion del Fork
            configFork.forEach((value, index) => {
                // Consultar la configuracion
                const confW = this.getConfDefinitionFork(value.fork);
                // Que elementos tienen una funcion para pasarla como parametro.
                if (value.fnc) {
                    // Buscar en el contexto actual.
                    const mtdExecute = value.fnc
                        .split('.')
                        .reduce((o, x) => {
                        return typeof o === 'undefined' || o === null
                            ? o
                            : o[x];
                    }, this);
                    // Ver si no existe
                    if (mtdExecute) {
                        // Asignar
                        value.fnc = mtdExecute.bind(this);
                    }
                    else {
                        this.logs.error(`No existe el metodo ${value.fnc} que se trata de asignar al fork ${value.fork}`);
                        process.exit(-1);
                    }
                }
                // Datos del Fork
                configFork[index] = Object.assign(value, {
                    conf: confW,
                });
            });
            // Regresar el valor
            return configFork;
        }
        else {
            return null;
        }
    }
    /**
     * Metodo para recuperar los datos de un Fork
     * @param name Nombre del fork definido
     */
    getConfDefinitionFork(name) {
        // Validar si existe
        if (this.ForksColl.has(name)) {
            return this.ForksColl.get(name);
        }
        else {
            this.logs.error(`No existe una configuracion par el fork ${name}`);
            process.exit(-1);
        }
    }
    /**
     * Metodo para recuperar los parametros requeridos
     * para un metodo.
     * @param name Nombre del metodo
     */
    getParams(name) {
        if (Reflect.hasMetadata(`${name}:conf`, this)) {
            // Recuperar el valor
            return Reflect.getMetadata(`${name}:conf`, this);
        }
        return [];
    }
    /**
     * Metodo para realizar la busqueda de una bifurcacion.
     * @param Col Coleccion de callbacks
     * @param find Bifurcacion que es buscada
     */
    existCallback(Coll, find) {
        // Buscar en todos los registros
        for (const cb of Coll) {
            // Validar si coincide con el elemento buscado
            if (cb.method === find.method && cb.name === find.name) {
                return true;
            }
        }
        return false;
    }
}
exports.Microservice = Microservice;
